#!/usr/bin/env python3
#
# author: scott olesen <swo@alum.mit.edu>

from __future__ import print_function
import argparse, sys, warnings
import pandas as pd, numpy as np
import Levenshtein
from Bio import SeqIO
import scipy.stats

class OTU:
    '''
    Object for keeping track of an OTU's distribution and computing genetic distances
    '''
    def __init__(self, name, sequence, counts):
        '''
        name: str
          OTU ID
        sequence: str
          OTU's nucleotide sequence
        counts: numpy.Array
          length of sequence should be the number of samples
        '''
        # make this assertion so that lists of counts don't get concatenated
        self.name = name
        self.sequence = sequence
        self.counts = np.array(counts)

        self.abundance = sum(self.counts)

    def __eq__(self, other):
        return self.name == other.name and self.sequence == other.sequence and all(self.counts == other.counts)

    def __repr__(self):
        return "OTU(name={}, sequence={}, counts={})".format(repr(self.name), repr(self.sequence), repr(self.counts))

    def absorb(self, other):
        '''
        Add another OTU's counts to this one

        other: OTU

        returns: nothing
        '''
        self.counts += other.counts
        self.abundance += other.abundance

    def distance_to(self, other):
        '''
        Length-adjusted Levenshtein "distance" to other OTU

        other: OTU
          distance to this OTU

        returns: float
        '''
        return Levenshtein.distance(self.sequence, other.sequence) / (0.5 * (len(self.sequence) + len(other.sequence)))

    @staticmethod
    def _D_helper(x):
        '''A helper function for the _D method'''
        x = np.array(x)
        x = x[x > 0]
        return np.sum(x * np.log(x)) - (np.sum(x) * np.log(np.sum(x)))

    @classmethod
    def _D(cls, x, y):
        '''
        Statistic for the likelihood ratio test. See docs for mathematical derivation.
        '''
        x = np.array(x)
        y = np.array(y)
        return -2.0 * (cls._D_helper(x + y) - cls._D_helper(x) - cls._D_helper(y))

    @classmethod
    def _distribution_test_pval(cls, x, y):
        '''
        P-value from the likelihood ratio test comparing the distribution of the abundances
        of two taxa (x and y). See docs for explanation of the test.
        '''
        assert len(x) == len(y)
        df = len(x) - 1
        return scipy.stats.chi2.sf(cls._D(x, y), df=df)

    def distribution_pval(self, other):
        '''
        P-value from the likelihood ratio test comparing the distribution of the abundances
        of two OTU objects. See docs for explanation of the test.

        other: OTU

        returns: float
        '''
        return self._distribution_test_pval(self.counts, other.counts)


class DBCaller:
    '''
    Object for processing the sequence table and distance matrix into an OTU table.
    '''
    def __init__(self, seq_table, records, max_dist, min_fold, threshold_pval, log=None):
        '''
        seq_table: pandas.DataFrame
          Samples on the columns; sequences on the rows
        records: index of Bio.Seq
          Indexed, unaligned input sequences. This could come from BioPython's
          SeqIO.to_dict or SeqIO.index.
        max_dist: float
          genetic distance cutoff above which a sequence will not be merged into an OTU
        min_fold: float
          Multiply the sequence's abundance by this fold to get the minimum abundance
          of an OTU for merging
        threshold_pval: float
          P-value below which a sequence will not be merged into an OTU
        log: filehandle
          Log file reporting the abundance, genetic, and distribution checks.
        '''
        self.seq_table = seq_table
        self.records = records
        self.max_dist = max_dist
        self.min_fold = min_fold
        self.threshold_pval = threshold_pval
        self.log = log

        # get a list of the names of the sequences in order of their (decreasing) abundance
        self.seq_abunds = self.seq_table.sum(axis=1).sort_values(ascending=False)

        # check that all sequence IDs in the table are in the fasta
        missing_ids = [seq_id for seq_id in self.seq_abunds.index if seq_id not in self.records]
        if len(missing_ids) > 0:
            raise RuntimeError("{} sequence IDs found in the sequence table but not in the fasta: {}".format(len(missing_ids), missing_ids))

        # initialize OTU information
        self.membership = {}
        self.otus = []

    def ga_matches(self, candidate):
        '''
        OTUs that meet the genetic and abundance criteria

        candidate: OTU
          sequence to evaluate
        '''

        # find abundance matches
        min_abundance = self.min_fold * candidate.abundance
        abundance_matches = [otu for otu in self.otus if otu.abundance > min_abundance]

        if self.log is not None:
            print(candidate.name, 'abundance_check', *[otu.name for otu in abundance_matches], sep='\t', file=self.log)

        if len(abundance_matches) == 0:
            return []
        else:
            # find genetic matches (in order of decreasing genetic distance)
            matches_distances = [(otu.distance_to(candidate), otu) for otu in abundance_matches]
            matches_distances.sort(key=lambda x: (x[0], -x[1].abundance, x[1].name))
            matches = [otu for dist, otu in matches_distances if dist < self.max_dist]

            if self.log is not None:
                print(candidate.name, 'genetic_check', *[otu.name for otu in matches], sep='\t', file=self.log)

            return matches

    def _process_record(self, record_id):
        '''
        Process the next sequence: run the genetic, abundance, and distribution checks, either
        merging the sequence into an existing OTU or creating a new OTU.
        '''
        assert record_id in self.seq_table.index
        record = self.records[record_id]

        candidate = OTU(record.id, str(record.seq), self.seq_table.loc[record.id])

        if self.log is not None:
            print('seq', candidate.name, sep='\t', file=self.log)

        merged = False
        for otu in self.ga_matches(candidate):
            test_pval = candidate.distribution_pval(otu)

            if self.log is not None:
                print(candidate.name, 'distribution_check', otu.name, test_pval, sep='\t', file=self.log)

            if test_pval > self.threshold_pval:
                otu.absorb(candidate)
                self.membership[otu.name].append(candidate.name)
                merged = True
                break

        if not merged:
            # form own otu
            self.otus.append(candidate)
            self.membership[candidate.name] = [candidate.name]

    def generate_otu_table(self):
        '''
        Process all the input sequences to make an OTU table.

        returns: pandas.DataFrame
          OTU table (which can also be found at instance.otu_table)
        '''
        for record_id in self.seq_abunds.index:
            self._process_record(record_id)

        self.otus.sort(key=lambda otu: otu.abundance, reverse=True)
        self.otu_table = pd.DataFrame([otu.counts for otu in self.otus], index=[otu.name for otu in self.otus])
        self.otu_table.columns = self.seq_table.columns

        return self.otu_table

    def write_otu_table(self, output):
        '''
        Write the QIIME-style OTU table to a file.

        output: filehandle
        '''
        self.otu_table.to_csv(output, sep='\t')

    def write_membership(self, output):
        '''
        Write the QIIME-style OTU mapping information to a file.

        output: filehandle
        '''
        for otu in self.otus:
            print(otu.name, *self.membership[otu.name], sep='\t', file=output)


def read_sequence_table(fn):
    '''
    Read in a table of sequences. The table must be tab-separated with exactly
    one header line of a field naming the sequences (e.g., "OTU", "OTU_ID",
    "seq", etc.) followed by tab-separated sample names. Sequence names are the
    first field of the following rows. The cells in the table are the counts of
    that sequence in that sample.

    fn: filename (or handle)

    returns: pandas.DataFrame
    '''

    # to ensure that the first column is read as a string,
    # the whole table must initially read as strings
    df = pd.read_table(fn, dtype={0: str}, header=0) # read in all columns as strings

    # BIOM format things will complain here
    if type(df.index) is pd.indexes.multi.MultiIndex:
        warnings.warn('Table was parsed with unusual indexes. Does this table comply with the tab-separated format?', RuntimeWarning)

    # TSV formats will complain here
    if df.shape[1] == 1:
        warnings.warn('Only one tab-separated column detected. Does this table comply with the tab-separated format?', RuntimeWarning)

    df.index = df.iloc[:,0] # specify the index as the first column
    df = df.iloc[:,1:].astype(int) # cast all data columns as integers
    return df

def call_otus(seq_table_fh, fasta_fh, output_fh, gen_crit, abund_crit, pval_crit, log=None, membership=None):
    '''
    Read in input files, call OTUs, and return output.

    seq_table_fh: filehandle
      sequence count table, tab-separated
    fasta_fh: filehandle or filename
      sequences fasta
    output_fh: filehandle
      place to write main output OTU table
    gen_crit, abund_crit, pval_crit: float
      threshold values for genetic criterion, abundance criterion, and distribution criterion (pvalue)
    log, membership: filehandles
      places to write supplementary output
    '''

    # ensure valid argument values
    assert gen_crit >= 0
    assert abund_crit >= 0.0
    assert pval_crit >= 0.0 and pval_crit <= 1.0

    # read in the sequences table
    seq_table = read_sequence_table(seq_table_fh)

    # set up the input fasta records
    records = SeqIO.index(fasta_fh, 'fasta')

    # generate the caller object
    caller = DBCaller(seq_table, records, gen_crit, abund_crit, pval_crit, log)
    caller.generate_otu_table()
    caller.write_otu_table(output_fh)

    if membership is not None:
        caller.write_membership(membership)

if __name__ == '__main__':
    p = argparse.ArgumentParser(description='dbOTU3: Distribution-based OTU calling')
    p.add_argument('table', type=argparse.FileType('r'), help='sequence count table')
    p.add_argument('fasta', help='sequences (trimmed if unpaired, merged if paired-end; always unaligned)')

    g = p.add_argument_group(title='criteria')
    g.add_argument('--dist', '-d', type=float, default=0.1, metavar='D', help='maximum genetic dissimilarity between sequences; more dissimilar sequence pairs do not pass the genetic criterion (default: 0.1)')
    g.add_argument('--abund', '-a', type=float, default=10.0, metavar='A', help='minimum fold difference for comparing two OTUs (0=no abundance criterion; default 10.0)')
    g.add_argument('--pval', '-p', type=float, default=0.0005, metavar='P', help='minimum p-value for merging OTUs (default: 0.0005)')

    g = p.add_argument_group(title='output options')
    g.add_argument('--output', '-o', default=sys.stdout, metavar='FILE', help='OTU table output (default: stdout)')
    g.add_argument('--membership', '-m', default=None, type=argparse.FileType('w'), metavar='FILE', help='QIIME-style OTU mapping file output')
    g.add_argument('--log', '-l', default=None, type=argparse.FileType('w'), metavar='FILE', help='log output')
    args = p.parse_args()

    call_otus(args.table, args.fasta, args.output, args.dist, args.abund, args.pval, log=args.log, membership=args.membership)



import pytest
from dbotu import read_sequence_table
from Bio import SeqIO
import os
try:
    from StringIO import StringIO
except:
    from io import StringIO



@pytest.fixture
def read_table():
    table_fh = StringIO('\n'.join(['\tS1\tS2\tS3', '001\t0\t1\t5', '002\t10\t25\t5', '003\t4\t0\t0', '004\t0\t0\t1']))
    return read_sequence_table(table_fh)

def test_shape(read_table):
    assert read_table.shape == (4,3)

def test_index(read_table):
    assert read_table.index[0] == '001'

def test_seqIDs(read_table):
    table = read_table
    fasta_fh = StringIO('\n'.join(['>001', 'ATCG', '>002', 'TCGA', '>003', 'AGCT', '>004', 'TTCC']))
    fasta_fh.seek(0)
    records = SeqIO.to_dict(SeqIO.parse(fasta_fh, 'fasta'))
    missing_ids = [seq_id for seq_id in table.index if seq_id not in records]
    assert len(missing_ids) == 0




from setuptools import find_packages, setup

with open('README.rst') as f:
    readme=f.read()

setup(name='dbotu',
    version='1.2.1',
    description='Distribution-based OTU calling',
    long_description=readme,
    author='Scott Olesen',
    author_email='swo@alum.mit.edu',
    platforms=['any'],
    license='License :: OSI Approved :: MIT License',
    url='http://github.com/swo/dbotu3',
    scripts=['dbotu.py'],
    install_requires=['numpy', 'pandas', 'scipy', 'biopython', 'python-Levenshtein'],
    package_date={'dbotu': ['data/input/counts.txt', 'data/input/seq.fa',
        'data/output/log.txt', 'data/output/membership.txt', 'data/output/otu.txt',
        'test/improper_data/counts.csv', 'test/improper_data/counts_biom.txt']},
    classifiers=['Development Status :: 4 - Beta', 'Environment :: Console',
        'Programming Language :: Python :: 2.7', 'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Bio-Informatics'],
    packages=find_packages()
    )



import pytest
from dbotu import *

import numpy as np, pandas as pd
import scipy.stats, scipy.optimize
from Bio import SeqIO
import os.path
# For Python 2 compatibility
try:
    from StringIO import StringIO
except:
    from io import StringIO


input_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', 'data', 'input')
fasta_fn = os.path.join(input_dir, 'seq.fa')
table_fn = os.path.join(input_dir, 'counts.txt')

@pytest.fixture
def caller():
    records = SeqIO.index(fasta_fn, 'fasta')
    table = read_sequence_table(table_fn)
    return DBCaller(table, records, max_dist=0.10, min_fold=10.0, threshold_pval=0.0005)

def test_init(caller):
    assert all(caller.seq_abunds[0:3] == pd.Series([2523, 1831, 1793], index=['seq9', 'seq0', 'seq2']))

def test_process_one(caller):
    caller._process_record(caller.seq_abunds.index[0])
    assert caller.otus[0] == OTU('seq9', 'GGAATATTGGTCAATGGAGGCAACTCTGAACCAGCCATGCCGCGTGCAGGATGACGGCCCTATGGGTTGTAAACTGCTTTTGTACCAGAGAAAACCCGAGTACGTGTACTCGGTTGATAGTATGGTAAGAATAAGCATCGGCTAACTTCGTGCCAGCAGCCGCGGTAAGACGAAGGATGCAAGCGTTATCCGGATTCATTGGGTTTAAAGGGTGCGTAGGCGGACCTGTAAGTCAGTGGTGAAATCTCTTTGCTTAACAAAGAAACTGCCATTGATACTGCAAGTCTAGAGTATAGATGACGTTGGCGGAATATGACATGTAGTGGTGAAATACTTAGATATGTCATAGAACACCGATTGCGAAGGCAGC', [947, 749, 274, 162, 391])

def test_process_until_first_merge(caller):
    for i in range(17):
        caller._process_record(caller.seq_abunds.index[i])

    seq9_counts = np.array([947, 749, 274, 162, 391])
    seq1537_counts = np.array([38, 25, 8, 5, 10])
    new_counts = seq9_counts + seq1537_counts

    assert caller.membership['seq9'] == ['seq9', 'seq1537']
    assert caller.otus[0] == OTU('seq9', 'GGAATATTGGTCAATGGAGGCAACTCTGAACCAGCCATGCCGCGTGCAGGATGACGGCCCTATGGGTTGTAAACTGCTTTTGTACCAGAGAAAACCCGAGTACGTGTACTCGGTTGATAGTATGGTAAGAATAAGCATCGGCTAACTTCGTGCCAGCAGCCGCGGTAAGACGAAGGATGCAAGCGTTATCCGGATTCATTGGGTTTAAAGGGTGCGTAGGCGGACCTGTAAGTCAGTGGTGAAATCTCTTTGCTTAACAAAGAAACTGCCATTGATACTGCAAGTCTAGAGTATAGATGACGTTGGCGGAATATGACATGTAGTGGTGAAATACTTAGATATGTCATAGAACACCGATTGCGAAGGCAGC', new_counts)

def test_full_process():
    records = SeqIO.index(fasta_fn, 'fasta')
    table = read_sequence_table(table_fn)

    table_fh = open(table_fn)

    test_otu_fh = StringIO()
    test_log_fh = StringIO()
    test_membership_fh = StringIO()

    call_otus(table_fh, fasta_fn, test_otu_fh, gen_crit=0.10, abund_crit=10.0, pval_crit=0.0005, log=test_log_fh, membership=test_membership_fh)

    test_log_content = test_log_fh.getvalue()
    test_otu_content = test_otu_fh.getvalue()
    test_membership_content = test_membership_fh.getvalue()

    output_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', 'data', 'output')
    otu_fn, log_fn, membership_fn = [os.path.join(output_dir, fn) for fn in ['otu.txt', 'log.txt', 'membership.txt']]

    with open(otu_fn) as f:
        otu_content = f.read()

    with open(log_fn) as f:
        log_content = f.read()

    with open(membership_fn) as f:
        membership_content = f.read()

    assert otu_content == test_otu_content
    assert membership_content == test_membership_content

    # I don't compare log_content because sometimes the last decimal place in
    # the numerical values is different. I could write a function that compares
    # the values in the log.
    #
    # assert log_content == test_log_content

improper_input_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'improper_data')
csv_table_fn = os.path.join(improper_input_dir, 'counts.csv')
biom_table_fn = os.path.join(improper_input_dir, 'counts_biom.txt')

def test_warn_csv_format():
    with pytest.warns(RuntimeWarning):
        read_sequence_table(csv_table_fn)

def test_warn_biom_format():
    with pytest.warns(RuntimeWarning):
        read_sequence_table(biom_table_fn)
